#ifndef PERFORMANCE_MONITOR_HPP_
#define PERFORMANCE_MONITOR_HPP_

#include <atomic>
#include <chrono>
#include <deque>
#include <mutex>
#include <string>
#include <unordered_map>
#include <vector>
#include <nlohmann/json.hpp>
#include "rclcpp/rclcpp.hpp"
#include "simple_timer/rt-sched.hpp"

namespace priority_executor {

enum class PerformanceEventType {
    CALLBACK_READY,
    CALLBACK_START,
    CALLBACK_END,
    DEADLINE_MISSED,
    DEADLINE_MET,
    CHAIN_START,
    CHAIN_END
};

struct PerformanceEvent {
    uint64_t timestamp;         // High-resolution timestamp (microseconds)
    PerformanceEventType type;  // Type of event
    std::string node_name;      // Name of the node
    std::string callback_name;  // Name of the callback
    int chain_id{-1};              // Chain identifier
    bool is_first_in_chain{false};    // First callback in chain
    bool is_last_in_chain{false};     // Last callback in chain
    uint64_t deadline{0};         // Deadline (if applicable)
    uint64_t processing_time{0};  // Time spent executing (if applicable)
    int executor_id{0};           // Executor identifier
    std::string additional_data; // JSON string for extensible data
    
    // Helper method to extract callback type from additional_data
    std::string get_callback_type() const {
        if (additional_data.empty()) return "unknown";
        
        try {
            nlohmann::json json = nlohmann::json::parse(additional_data);
            if (json.contains("callback_type")) {
                return json["callback_type"].get<std::string>();
            }
        } catch (...) {}
        
        return "unknown";
    }
    
    // Helper method to extract thread info from additional_data
    int get_thread_id() const {
        if (additional_data.empty()) return -1;
        
        try {
            nlohmann::json json = nlohmann::json::parse(additional_data);
            if (json.contains("thread_info") && json["thread_info"].is_object() && 
                json["thread_info"].contains("thread_id")) {
                return json["thread_info"]["thread_id"].get<int>();
            }
            // Direct thread_id if present
            if (json.contains("thread_id")) {
                return json["thread_id"].get<int>();
            }
        } catch (...) {}
        
        return -1;
    }
    
    // Helper method to create a descriptive identifier for this callback
    std::string get_descriptive_id() const {
        std::stringstream ss;
        ss << node_name;
        
        if (!callback_name.empty()) {
            ss << "/" << callback_name;
        }
        
        if (chain_id >= 0) {
            ss << " (chain:" << chain_id;
            if (is_first_in_chain) ss << " start";
            if (is_last_in_chain) ss << " end";
            ss << ")";
        }
        
        return ss.str();
    }
    
    // Convert additional_data to JSON object
    nlohmann::json get_additional_data_json() const {
        try {
            if (!additional_data.empty()) {
                return nlohmann::json::parse(additional_data);
            }
        } catch (...) {}
        
        return nlohmann::json::object();
    }
};

class PerformanceMonitor {
public:
    static PerformanceMonitor& getInstance();
    
    void recordEvent(const PerformanceEvent& event);
    void setBufferSize(size_t size);
    void setAutoDumpThreshold(size_t threshold);
    void setDumpPath(const std::string& path);
    bool dumpToFile(const std::string& filename);
    void enable(bool enabled = true);
    void clear();
    
    // Helper method to set monitoring options in one call
    void setMonitoringOptions(size_t buffer_size, size_t auto_dump_threshold, 
                              const std::string& dump_path) {
        setBufferSize(buffer_size);
        setAutoDumpThreshold(auto_dump_threshold);
        setDumpPath(dump_path);
    }
    
    // Get all events for a specific callback
    std::vector<PerformanceEvent> getEventsForCallback(const std::string& node_name, 
                                                      const std::string& callback_name) const {
        std::vector<PerformanceEvent> result;
        std::lock_guard<std::mutex> lock(event_mutex_);
        
        for (const auto& event : event_buffer_) {
            if (event.node_name == node_name && event.callback_name == callback_name) {
                result.push_back(event);
            }
        }
        
        return result;
    }
    
    // Get all events for a specific chain
    std::vector<PerformanceEvent> getEventsForChain(int chain_id) const {
        std::vector<PerformanceEvent> result;
        std::lock_guard<std::mutex> lock(event_mutex_);
        
        for (const auto& event : event_buffer_) {
            if (event.chain_id == chain_id) {
                result.push_back(event);
            }
        }
        
        return result;
    }
    
    // Get all events by type
    std::vector<PerformanceEvent> getEventsByType(PerformanceEventType type) const {
        std::vector<PerformanceEvent> result;
        std::lock_guard<std::mutex> lock(event_mutex_);
        
        for (const auto& event : event_buffer_) {
            if (event.type == type) {
                result.push_back(event);
            }
        }
        
        return result;
    }
    
    // Get a formatted report of callbacks and their execution times
    std::string getCallbackExecutionReport() const {
        std::lock_guard<std::mutex> lock(event_mutex_);
        std::stringstream ss;
        
        ss << "Callback Execution Report:\n";
        ss << "=========================\n\n";
        
        // Maps to store total execution time and count per callback
        std::unordered_map<std::string, uint64_t> total_time;
        std::unordered_map<std::string, int> call_count;
        std::unordered_map<std::string, uint64_t> max_time;
        std::unordered_map<std::string, uint64_t> min_time;
        
        // Process all CALLBACK_END events (they contain processing_time)
        for (const auto& event : event_buffer_) {
            if (event.type == PerformanceEventType::CALLBACK_END && event.processing_time > 0) {
                std::string callback_id = event.get_descriptive_id();
                total_time[callback_id] += event.processing_time;
                call_count[callback_id]++;
                max_time[callback_id] = std::max(max_time[callback_id], event.processing_time);
                if (min_time[callback_id] == 0 || event.processing_time < min_time[callback_id]) {
                    min_time[callback_id] = event.processing_time;
                }
            }
        }
        
        // Print results sorted by callback id
        std::vector<std::string> callback_ids;
        for (const auto& entry : call_count) {
            callback_ids.push_back(entry.first);
        }
        
        std::sort(callback_ids.begin(), callback_ids.end());
        
        for (const auto& id : callback_ids) {
            double avg_time = static_cast<double>(total_time[id]) / call_count[id];
            ss << id << "\n";
            ss << "  Calls: " << call_count[id] << "\n";
            ss << "  Avg time: " << avg_time << " µs\n";
            ss << "  Min time: " << min_time[id] << " µs\n";
            ss << "  Max time: " << max_time[id] << " µs\n";
            ss << "  Total time: " << total_time[id] << " µs\n\n";
        }
        
        return ss.str();
    }

private:
    PerformanceMonitor();
    ~PerformanceMonitor() = default;
    
    PerformanceMonitor(const PerformanceMonitor&) = delete;
    PerformanceMonitor& operator=(const PerformanceMonitor&) = delete;
    PerformanceMonitor(PerformanceMonitor&&) = delete;
    PerformanceMonitor& operator=(PerformanceMonitor&&) = delete;

    mutable std::mutex event_mutex_;
    std::vector<PerformanceEvent> event_buffer_;
    size_t buffer_size_{10000};
    size_t auto_dump_threshold_{0};
    std::string dump_path_{"performance_logs"};
    std::atomic<bool> enabled_{true};
    
    std::string eventsToJson() const;
    void checkAndAutoDump();
};

}  // namespace priority_executor

#endif  // PERFORMANCE_MONITOR_HPP_