#include "priority_executor/multithread_priority_executor.hpp"
#include "priority_executor/performance_monitor.hpp"

#include <chrono>
#include <functional>
#include <memory>
#include <mutex>
#include <sstream>
#include <thread>
#include <unordered_map>
#include <nlohmann/json.hpp>

#include "rcpputils/scope_exit.hpp"
#include "rclcpp/any_executable.hpp"
#include "simple_timer/rt-sched.hpp"

namespace priority_executor {

std::unordered_map<MultithreadTimedExecutor*,
                   std::shared_ptr<rclcpp::detail::MutexTwoPriorities>>
    MultithreadTimedExecutor::wait_mutex_set_;
std::mutex MultithreadTimedExecutor::shared_wait_mutex_;

MultithreadTimedExecutor::MultithreadTimedExecutor(
    rclcpp::ExecutorOptions const &options,
    std::string const &name,
    int number_of_threads,
    std::chrono::nanoseconds next_exec_timeout)
    : TimedExecutor(options, name) {
    std::lock_guard<std::mutex> wait_lock(MultithreadTimedExecutor::shared_wait_mutex_);
    wait_mutex_set_[this] = std::make_shared<rclcpp::detail::MutexTwoPriorities>();
    number_of_threads_ = number_of_threads;
    next_exec_timeout_ = next_exec_timeout;
    logger_ = create_logger();
}

size_t MultithreadTimedExecutor::get_number_of_threads()
{
    return number_of_threads_;
}

void MultithreadTimedExecutor::spin() {
    if (spinning.exchange(true)) {
        throw std::runtime_error("spin() called while already spinning");
    }

    std::vector<std::thread> threads;
    size_t thread_id = 0;
    {
        auto wait_mutex = MultithreadTimedExecutor::wait_mutex_set_[this];
        auto low_priority_wait_mutex = wait_mutex->get_low_priority_lockable();
        std::lock_guard<rclcpp::detail::MutexTwoPriorities::LowPriorityLockable>
            wait_lock(low_priority_wait_mutex);
        for (; thread_id < number_of_threads_ - 1; ++thread_id) {
            auto const func = std::bind(&MultithreadTimedExecutor::run, this, thread_id);
            threads.emplace_back(func);
        }
    }
    run(thread_id);
    for (auto& thread : threads) {
        thread.join();
    }
    spinning.store(false);
}

void MultithreadTimedExecutor::run(size_t _thread_number) {
    // set affinity
    cpu_set_t cpuset;
    CPU_ZERO(&cpuset);
    CPU_SET(_thread_number, &cpuset);
    int rc = pthread_setaffinity_np(pthread_self(), sizeof(cpu_set_t), &cpuset);
    if (rc != 0)
    {
        RCLCPP_ERROR(rclcpp::get_logger("priority_executor"),
                     "Error calling pthread_setaffinity_np: %d", rc);
    }

    // Create thread context information for logs
    nlohmann::json thread_info;
    thread_info["thread_id"] = _thread_number;
    thread_info["cpu_affinity"] = _thread_number;
    std::string thread_info_str = thread_info.dump();

    while (rclcpp::ok(this->context_) && spinning.load()) {
        rclcpp::AnyExecutable any_exec;
        {
            auto wait_mutex = MultithreadTimedExecutor::wait_mutex_set_[this];
            auto low_priority_wait_mutex = wait_mutex->get_low_priority_lockable();
            std::lock_guard<rclcpp::detail::MutexTwoPriorities::LowPriorityLockable>
                wait_lock(low_priority_wait_mutex);
            if (!rclcpp::ok(this->context_) || !spinning.load()) {
                return;
            }
            if (!get_next_executable(any_exec, next_exec_timeout_)) {
                continue;
            }
            if (any_exec.timer) {
                // Guard against multiple threads getting the same timer.
                if (scheduled_timers_.count(any_exec.timer) != 0) {
                    // Make sure that any_exec's callback group is reset before
                    // the lock is released.
                    if (any_exec.callback_group) {
                        any_exec.callback_group->can_be_taken_from().store(true);
                    }
                    continue;
                }
                scheduled_timers_.insert(any_exec.timer);
            }
        }

        // Record execution start with thread info
        auto start = std::chrono::steady_clock::now();
        uint64_t start_time = std::chrono::duration_cast<std::chrono::microseconds>(
            start.time_since_epoch()).count();
            
        PerformanceEvent start_event;
        start_event.timestamp = start_time;
        start_event.type = PerformanceEventType::CALLBACK_START;
        start_event.executor_id = getExecutorId();
        
        if (any_exec.node_base) {
            start_event.node_name = any_exec.node_base->get_name();
            
            // Enhanced callback identification based on executable type
            if (any_exec.timer) {
                // For timer callbacks
                std::stringstream ss;

                // retrieve the timer period
                int64_t period;
                rcl_timer_get_period(any_exec.timer->get_timer_handle().get(), &period);

                // convert period (which is in nanoseconds) to milliseconds
                std::chrono::nanoseconds period_ns(period);
                std::chrono::milliseconds period_ms = std::chrono::duration_cast<std::chrono::milliseconds>(period_ns);

                ss << "timer_" << period_ms.count() << "ms";
                start_event.callback_name = ss.str();
                
                // Add memory address as unique identifier
                std::stringstream addr;
                addr << "@" << any_exec.timer.get();
                nlohmann::json additional;
                additional["callback_type"] = "timer";
                additional["timer_addr"] = addr.str();
                additional["thread_info"] = thread_info;
                start_event.additional_data = additional.dump();
            } 
            else if (any_exec.subscription) {
                // For subscription callbacks
                start_event.callback_name = any_exec.subscription->get_topic_name();
                
                nlohmann::json additional;
                additional["callback_type"] = "subscription";
                additional["thread_info"] = thread_info;
                start_event.additional_data = additional.dump();
            }
            else if (any_exec.service) {
                start_event.callback_name = any_exec.service->get_service_name();
                
                nlohmann::json additional;
                additional["callback_type"] = "service";
                additional["thread_info"] = thread_info;
                start_event.additional_data = additional.dump();
            }
            else if (any_exec.client) {
                start_event.callback_name = "client_callback";
                
                nlohmann::json additional;
                additional["callback_type"] = "client";
                additional["thread_info"] = thread_info;
                start_event.additional_data = additional.dump();
            }
            else if (any_exec.waitable) {
                start_event.callback_name = "waitable_callback";
                
                nlohmann::json additional;
                additional["callback_type"] = "waitable";
                additional["thread_info"] = thread_info;
                start_event.additional_data = additional.dump();
            }
            
            // Add chain information
            std::shared_ptr<PriorityExecutable> executable = nullptr;
            if (prio_memory_strategy_) {
                if (any_exec.timer) {
                    executable = prio_memory_strategy_->get_priority_settings(
                        any_exec.timer->get_timer_handle());
                } else if (any_exec.subscription) {
                    executable = prio_memory_strategy_->get_priority_settings(
                        any_exec.subscription->get_subscription_handle());
                }
            }

            if (executable) {
                start_event.chain_id = executable->chain_id;
                start_event.is_first_in_chain = executable->is_first_in_chain;
                start_event.is_last_in_chain = executable->is_last_in_chain;
                
                if (executable->deadlines && !executable->deadlines->empty()) {
                    start_event.deadline = executable->deadlines->front();
                }
                
                // Include the priority information
                try {
                    nlohmann::json additional = start_event.additional_data.empty() ? 
                        nlohmann::json() : nlohmann::json::parse(start_event.additional_data);
                        
                    additional["priority"] = executable->priority;
                    
                    if (executable->chain_id >= 0) {
                        additional["chain_id"] = executable->chain_id;
                    }
                    
                    if (executable->sched_type != ExecutableScheduleType::DEFAULT) {
                        additional["schedule_type"] = static_cast<int>(executable->sched_type);
                    }
                    
                    start_event.additional_data = additional.dump();
                } catch (...) {
                    // If there was an error parsing the JSON, create new JSON object
                    nlohmann::json additional;
                    additional["priority"] = executable->priority;
                    additional["thread_info"] = thread_info;
                    start_event.additional_data = additional.dump();
                }
            }
        } else {
            start_event.additional_data = thread_info_str;
        }
        
        PerformanceMonitor::getInstance().recordEvent(start_event);

        // Execute the callback
        execute_any_executable(any_exec);

        // Record execution completion
        auto end = std::chrono::steady_clock::now();
        uint64_t end_time = std::chrono::duration_cast<std::chrono::microseconds>(
            end.time_since_epoch()).count();
        uint64_t processing_time = end_time - start_time;

        // Create identical event structure but for end event
        PerformanceEvent end_event = start_event;
        end_event.timestamp = end_time; 
        end_event.type = PerformanceEventType::CALLBACK_END;
        end_event.processing_time = processing_time;
        
        // Update the additional data to include processing time
        try {
            nlohmann::json additional = end_event.additional_data.empty() ? 
                nlohmann::json() : nlohmann::json::parse(end_event.additional_data);
            additional["processing_time_us"] = processing_time;
            end_event.additional_data = additional.dump();
        } catch (...) {
            // If JSON parsing failed, create a new object
            nlohmann::json additional;
            additional["processing_time_us"] = processing_time;
            additional["thread_info"] = thread_info;
            end_event.additional_data = additional.dump();
        }
        
        PerformanceMonitor::getInstance().recordEvent(end_event);

        if (any_exec.timer) {
            auto wait_mutex = MultithreadTimedExecutor::wait_mutex_set_[this];
            auto high_priority_wait_mutex = wait_mutex->get_high_priority_lockable();
            std::lock_guard<rclcpp::detail::MutexTwoPriorities::HighPriorityLockable>
                wait_lock(high_priority_wait_mutex);
            auto const it = scheduled_timers_.find(any_exec.timer);
            if (it != scheduled_timers_.end()) {
                scheduled_timers_.erase(it);
            }
        }
        // Clear the callback_group to prevent the AnyExecutable destructor from
        // resetting the callback group `can_be_taken_from`
        any_exec.callback_group.reset();
        if (prio_memory_strategy_ != nullptr)
        {
            auto wait_mutex = MultithreadTimedExecutor::wait_mutex_set_[this];
            auto low_priority_wait_mutex = wait_mutex->get_low_priority_lockable();
            std::lock_guard<rclcpp::detail::MutexTwoPriorities::LowPriorityLockable> wait_lock(low_priority_wait_mutex);
            prio_memory_strategy_->post_execute(any_exec, _thread_number);
        }
    }
}
}  // namespace priority_executor
