#include "priority_executor/performance_monitor.hpp"
#include <filesystem>
#include <fstream>
#include <nlohmann/json.hpp>
#include <sstream>
#include <iomanip>

namespace priority_executor {

PerformanceMonitor& PerformanceMonitor::getInstance() {
    static PerformanceMonitor instance;
    return instance;
}

PerformanceMonitor::PerformanceMonitor() {
    event_buffer_.reserve(buffer_size_);
    std::filesystem::create_directories(dump_path_);
}

void PerformanceMonitor::recordEvent(const PerformanceEvent& event) {
    if (!enabled_) return;
    
    {
        std::lock_guard<std::mutex> lock(event_mutex_);
        event_buffer_.push_back(event);
        
        if (auto_dump_threshold_ > 0 && event_buffer_.size() >= auto_dump_threshold_) {
            checkAndAutoDump();
        }
    }
}

void PerformanceMonitor::setBufferSize(size_t size) {
    std::lock_guard<std::mutex> lock(event_mutex_);
    buffer_size_ = size;
    event_buffer_.reserve(size);
}

void PerformanceMonitor::setAutoDumpThreshold(size_t threshold) {
    auto_dump_threshold_ = threshold;
}

void PerformanceMonitor::setDumpPath(const std::string& path) {
    dump_path_ = path;
    std::filesystem::create_directories(dump_path_);
}

void PerformanceMonitor::enable(bool enabled) {
    enabled_ = enabled;
}

void PerformanceMonitor::clear() {
    std::lock_guard<std::mutex> lock(event_mutex_);
    event_buffer_.clear();
}

std::string PerformanceMonitor::eventsToJson() const {
    nlohmann::json root = nlohmann::json::array();
    
    for (const auto& event : event_buffer_) {
        nlohmann::json event_json;
        
        event_json["timestamp"] = event.timestamp;
        
        switch (event.type) {
            case PerformanceEventType::CALLBACK_READY:
                event_json["type"] = "callback_ready";
                break;
            case PerformanceEventType::CALLBACK_START:
                event_json["type"] = "callback_start";
                break;
            case PerformanceEventType::CALLBACK_END:
                event_json["type"] = "callback_end";
                break;
            case PerformanceEventType::DEADLINE_MISSED:
                event_json["type"] = "deadline_missed";
                break;
            case PerformanceEventType::DEADLINE_MET:
                event_json["type"] = "deadline_met";
                break;
            case PerformanceEventType::CHAIN_START:
                event_json["type"] = "chain_start";
                break;
            case PerformanceEventType::CHAIN_END:
                event_json["type"] = "chain_end";
                break;
        }
        
        event_json["node_name"] = event.node_name;
        event_json["callback_name"] = event.callback_name;
        event_json["chain_id"] = event.chain_id;
        event_json["is_first_in_chain"] = event.is_first_in_chain;
        event_json["is_last_in_chain"] = event.is_last_in_chain;
        
        if (event.deadline > 0) {
            event_json["deadline"] = event.deadline;
        }
        
        if (event.processing_time > 0) {
            event_json["processing_time"] = event.processing_time;
        }
        
        event_json["executor_id"] = event.executor_id;
        
        if (!event.additional_data.empty()) {
            try {
                event_json["additional_data"] = nlohmann::json::parse(event.additional_data);
            } catch (...) {
                event_json["additional_data"] = event.additional_data;
            }
        }
        
        root.push_back(event_json);
    }
    
    return root.dump(2);  // Indent with 2 spaces
}

bool PerformanceMonitor::dumpToFile(const std::string& filename) {
    std::lock_guard<std::mutex> lock(event_mutex_);
    
    if (event_buffer_.empty()) {
        return true;  // Nothing to dump
    }
    
    std::string full_path = dump_path_ + "/" + filename;
    std::ofstream file(full_path);
    if (!file.is_open()) {
        return false;
    }
    
    file << eventsToJson();
    event_buffer_.clear();
    return true;
}

void PerformanceMonitor::checkAndAutoDump() {
    if (event_buffer_.size() >= auto_dump_threshold_) {
        auto now = std::chrono::system_clock::now();
        auto timestamp = std::chrono::duration_cast<std::chrono::milliseconds>(
            now.time_since_epoch()).count();
            
        std::stringstream ss;
        ss << "performance_log_" << timestamp << ".json";
        dumpToFile(ss.str());
    }
}

}  // namespace priority_executor